﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Proteus;
using Microsoft.Win32;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;
using System.IO;
using IWshRuntimeLibrary;
namespace Spark.Installer
{

    /*
     * Installer File format
     * [exe data]
     * [Token string] [string]
     * [Options Table][int8 (count)][ InstallerOption...]
     * [File Table][int8(count)][InstallerFile...]
     * [Files...] [raw data]
     * 
     */
    public class InstallManager
    {
        #region Private:Members
        private InstallOptions _objOptions;
        private Object _objOptionsLockObject = new Object();
        //private InstallerFileTable _objFileTable;
        private double _dblProgress = 0;//Primitives are atomic
        private CancellationTokenSource _installerCancellationToken;
        private Object _objInstallErrorsLockObject = new Object();
        private List<string> _lstInstallErrors = new List<string>();
        private string _strCurrentFile;
        private InstallState _enumInstallState = InstallState.Pending;
        private bool _blnUserCancelled = false; // set if user has pressed cancel during install
        #endregion

        #region Public:Properties
        public bool UserCancelled
        {
            get
            {
                return _blnUserCancelled;
            }
            private set
            {
                _blnUserCancelled = value;
            }
        }
        public InstallOptions Options
        {
            get
            {
                lock (_objOptionsLockObject)
                    return _objOptions;
            }
            private set
            {
                lock (_objOptionsLockObject)
                    _objOptions = value;
            }
        }
        public double InstallProgress
        {
            get
            {
                return _dblProgress;
            }
            private set { }
        }
        public string CurrentFile
        {
            get
            {
                return _strCurrentFile;
            }
            private set { }
        }
        public List<string> InstallErrors
        {
            get
            {
                lock (_objInstallErrorsLockObject)
                    return _lstInstallErrors;
            }
            private set
            {
                lock (_objInstallErrorsLockObject)
                    _lstInstallErrors = value;
            }
        }
        public InstallState InstallState
        {
            get
            {
                return _enumInstallState;
            }
            private set
            {
                _enumInstallState = value;
            }
        }
        #endregion

        #region Public:Methods
        public List<ShortcutOption> GetShortcutOptions()
        {
            return _objOptions.GetShortcutOptions();
        }

        public string ReplaceValueVariables(string strValue)
        {
            if (strValue.Contains(InstallerVariable.ProgramFiles))
            {
                //**Also - Environment.GetEnvironmentVariable("ProgramFiles"); - returns x86 program files.
                string strProgramFiles = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles);
                strValue = strValue.Replace(InstallerVariable.ProgramFiles, strProgramFiles);
            }
            if (strValue.Contains(InstallerVariable.ProgramFilesX86))
            {
                string strProgramFiles = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFilesX86);
                strValue = strValue.Replace(InstallerVariable.ProgramFiles, strProgramFiles);
            }
            if (strValue.Contains(InstallerVariable.Desktop))
            {
                string strProgramFiles = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
                strValue = strValue.Replace(InstallerVariable.ProgramFiles, strProgramFiles);
            }
            if (strValue.Contains(InstallerVariable.SystemRoot))
            {
                string strProgramFiles = Environment.GetFolderPath(Environment.SpecialFolder.Windows);
                strValue = strValue.Replace(InstallerVariable.ProgramFiles, strProgramFiles);
            }
            if (strValue.Contains(InstallerVariable.InstallDir))
            {
                InstallOption opt = GetOption(InstallOption.InstallLocation);
                if (opt == null)
                {
                    string strMsg = "Could not get install location.  Doesn't seem to have been set. Note: If you " +
                        " are running the INSTALLER then this will throw " +
                        " and your uninstall won't work.  Make sure to run the UNISNTALL program. (0101984)";
                    if (InstallState == InstallState.Pending)
                        Globals.Logger.LogWarn(strMsg);// This is okay when we first load the options and we aren't installing. However after user selects a path it MUST be set.
                    else
                        Globals.Throw(strMsg);
                }
                else
                {
                    strValue = strValue.Replace(InstallerVariable.InstallDir, opt.Value);
                }
            }
            if (strValue.Contains(InstallerVariable.UserStartup))
            {
                string strStartup = Environment.GetFolderPath(Environment.SpecialFolder.Startup);
                strValue = strValue.Replace(InstallerVariable.ProgramFiles, strStartup);
            }

            return strValue;
        }
        public List<InstallOption> GetAllOptionsForSection(InstallerSectionType sect)
        {
            return _objOptions.GetAllOptionsForSection(sect);
        }
        public InstallOption GetOption(string key)
        {
            return _objOptions.GetUniqueOption(key);
        }
        public InstallOption AddOrReplaceOption(string key, string value, InstallerSectionType sectionType)
        {
            InstallOption opt = new InstallOption(key, value, sectionType);
            _objOptions.AddOrReplaceOption(opt);
            return opt;
        }
        public void BeginInstallation()
        {
            Globals.Logger.LogInfo("Beginning install from main thread.");
            _installerCancellationToken = new CancellationTokenSource();
            CancellationToken ct = _installerCancellationToken.Token;

            Task.Factory.StartNew(new Action(ExecuteInstallation), _installerCancellationToken.Token);
        }
        public void BeginUninstall()
        {
            Globals.Logger.LogInfo("Beginning uninstall from main thread.");
            _installerCancellationToken = new CancellationTokenSource();
            CancellationToken ct = _installerCancellationToken.Token;

            Task.Factory.StartNew(new Action(ExecuteUninstall), _installerCancellationToken.Token);
        }
        public void UserCancelInstallation()
        {
            //User cancel only.  This allows us to skip the error screen.
            _installerCancellationToken.Cancel();
            UserCancelled = true;
        }
        public void LoadOptions()
        {
            Globals.Logger.LogInfo("Extracting Install Options");

            //Use a dummy to read past this installer exe binary, then read the options table.
            InstallerBinary dummy = new InstallerBinary();
            _objOptions = new InstallOptions();

            try
            {
                dummy.OpenStream();
                _objOptions.Deserialize(dummy, this);
            }
            catch (Exception ex)
            {
                Globals.Logger.LogInfo("Exception opening install options: " + ex.ToString());

                throw ex;
            }
            finally
            {
                dummy.CloseStream();
            }

            dummy = null;
        }
        public static bool ApplicationIsInstalled(ProgramInfo objProgramInfo)
        {
            bool blnValue = true;
            RegistryKey softwareKey = Registry.LocalMachine.OpenSubKey(@"Software", true);
            if (softwareKey == null)
                blnValue = false;
            else
            {
                RegistryKey publisherKey;
                publisherKey = softwareKey.OpenSubKey(objProgramInfo.Publisher, true);
                if (publisherKey == null)
                    blnValue = false;
                else
                {
                    RegistryKey appKey;
                    appKey = publisherKey.OpenSubKey(objProgramInfo.DisplayName, true);
                    if (appKey == null)
                        blnValue = false;
                    else
                        appKey.Close();

                    publisherKey.Close();
                }
                softwareKey.Close();
            }
            return blnValue;
        }

        #endregion

        #region Private:Methods

        private void ExecuteInstallation()
        {
            //** Main installation routine.
            DisplayMsg("Beginning Installation");
            InstallState = InstallState.Installing;
            InstallerFileTable objTable = null;

            string strInstallRoot = CreateInstallRoot();
            if (string.IsNullOrEmpty(strInstallRoot))
                return; // error

            double dblProgressBeforeCopy = 0;

            try
            {
                string strSavedDir = System.IO.Directory.GetCurrentDirectory();
                System.IO.Directory.SetCurrentDirectory(strInstallRoot);
                {
                    DisplayMsg("Executing Pre-Install");
                    ExecutePreInstall();

                    DisplayMsg("Creating install dir " + strInstallRoot);
                    if (!System.IO.Directory.Exists(strInstallRoot))
                        System.IO.Directory.CreateDirectory(strInstallRoot);

                    // - Do work
                    InstallerBinary objBinary;
                    DisplayMsg("Creating binary");
                    objBinary = new InstallerBinary();
                    objBinary.OpenStream();

                    //**Read past the install options section.  Note this
                    // is NOT used
                    DisplayMsg("Reading Dummy Options");
                    InstallOptions dummy = new InstallOptions();
                    dummy.Deserialize(objBinary, this);

                    //Validate appliation GUID for uninstall.
                    InstallOption guidOpt = GetOption(InstallOption.AppGuid);
                    if (guidOpt == null)
                        Globals.Throw("Could not install.  Appliation GUID was not specified in install options.");
                    else
                        ProgramInfo.TryParseGuidToUninstallGuid(guidOpt.Value);
                    
                    // - reload the options so we can set the correct install path (user selected)
                    _objOptions.Reload(this);

                    DisplayMsg("Reading File Table");
                    objTable = new InstallerFileTable(this);
                    objTable.Deserialize(objBinary);

                    _dblProgress = dblProgressBeforeCopy = 0.5;

                    DisplayMsg("Unpacking Files");
                    objTable.UnpackFiles(
                        objBinary,
                        strInstallRoot,
                        ref _dblProgress,
                        _installerCancellationToken.Token,
                        ref _strCurrentFile
                        );

                    objBinary.CloseStream();

                    DisplayMsg("Creating App Reg keys");
                    CreateAppRegistryKeys(_objOptions.GetProgramInfo(), strInstallRoot);

                    DisplayMsg("Creating Uninstaller");
                    CreateUninstaller(_objOptions.GetProgramInfo(), objBinary, objTable, strInstallRoot);

                    DisplayMsg("Creating Shortcuts");
                    CreateShortcuts(_objOptions.GetProgramInfo(), _objOptions);

                    DisplayMsg("Executing Post-Install");
                    ExecutePostInstall();
                }
                System.IO.Directory.SetCurrentDirectory(strSavedDir);

            }
            catch (Exception ex)
            {
                InstallState = InstallState.Canceled;
                InstallErrors.Add("Error: " + ex.ToString());
                Globals.Logger.LogError(ex.ToString(), false, true);
                RollbackOrUninstall(objTable, strInstallRoot, dblProgressBeforeCopy);
            }
            finally
            {
                EndInstallation();
            }
        }
        private void ExecutePreInstall()
        {
            ExecuteOptionsForSection(InstallerSectionType.PreInstallSection);
            DisplayMsg("PreInstall Complete");
        }
        private void ExecutePostInstall()
        {
            ExecuteOptionsForSection(InstallerSectionType.PostInstallSection);
            DisplayMsg("PostInstall Complete");
        }
        private void ExecutePreUninstall()
        {
            ExecuteOptionsForSection(InstallerSectionType.PreUninstallSection);
            DisplayMsg("PreUninstall Complete");
        }
        private void ExecutePostUninstall()
        {
            ExecuteOptionsForSection(InstallerSectionType.PostUninstallSection);
            DisplayMsg("PostUninstall Complete");
        }
        private void CheckThrowIfProcessCancelled()
        {
            if (_installerCancellationToken.Token.IsCancellationRequested)
                _installerCancellationToken.Token.ThrowIfCancellationRequested();
        }

        #region Install Options
        private void ExecuteOptionsForSection(InstallerSectionType type)
        {
            List<InstallOption> lst = GetAllOptionsForSection(type);
            foreach (InstallOption opt in lst)
            {
                try
                {
                    ExecuteSectionOption(opt);
                }
                catch (Exception ex)
                {
                    string strErr = "Exception thrown executing install/uninstall option:\r\n\r\n" + ex.ToString();
                    Globals.Logger.LogError(strErr);
                    DisplayMsg(strErr);
                    Exception ex2 = new Exception(strErr, ex);
                    throw ex2;
                }
            }
        }
        private void ExecuteSectionOption(InstallOption opt)
        {
            if (opt.Key.Equals(InstallOption.ExecuteSync))
            {
                ExecuteRunCommandOption(opt.Value);
            }
            else if (opt.Key.Equals(InstallOption.DeleteFile))
            {
                ExecuteDeleteFileOption(opt.Value);
            }
            else
            {
                Globals.Logger.LogError("FAILURE:, Install Option '" + opt.Key + "' was not recognized. Note that install options are case sensitive.");
            }
        }
        private void ExecuteRunCommandOption(string strPath)
        {
            strPath = ReplaceValueVariables(strPath);
            DisplayMsg("Executing: " + strPath);
            if (!System.IO.File.Exists(strPath))
            {
                string strerr = "Error: executable file " + strPath + " does not exist.";
                DisplayMsg(strerr);
                Globals.Logger.LogError(strerr);
            }
            else
            {
                try
                {
                    Globals.Logger.LogInfo("Current Directory: " + System.IO.Directory.GetCurrentDirectory());
                    ConsoleProcess _objConsoleProcess = new ConsoleProcess(true, false);
                    _objConsoleProcess.BeginAsync();
                    string cmdText = "cmd.exe /c ";
                    cmdText += StringUtils.DoubleEnquote(strPath) + " ";
                    _objConsoleProcess.Execute(cmdText, 0, _installerCancellationToken.Token);
                    //Begin/end exist here cbecuause romulus must exit so we knwo the makefile wa smade
                    _objConsoleProcess.Exit(true, 0);

                    //strPath = StringUtils.DoubleEnquote(strPath);
                    //System.Diagnostics.Process p = new System.Diagnostics.Process();
                    //p.StartInfo.FileName = "cmd.exe";
                    //p.StartInfo.Arguments = "/c " + strPath;
                    //p.StartInfo.CreateNoWindow = true;
                    //p.StartInfo.UseShellExecute = false;
                    //p.StartInfo.RedirectStandardError = true;
                    //p.StartInfo.RedirectStandardOutput = true;
                    //p.Start();
                    //WaitExitOrThrow(p);
                    //string output = p.StandardOutput.ReadToEnd();
                    //string error = p.StandardError.ReadToEnd();
                    DisplayMsg(_objConsoleProcess.CommandOutput + "\r\n");
                    //p.Close();
                }
                catch (Exception ex)
                {
                    string strErr = "External program " + strPath + " threw an exception:\r\n\r\n" + ex.ToString();
                    DisplayMsg(strErr);
                    Globals.Logger.LogError(strErr);
                    Exception ex2 = new Exception(strErr, ex);
                    throw ex2;
                }
            }
        }
        private void ExecuteDeleteFileOption(string strFilePath)
        {
            strFilePath = ReplaceValueVariables(strFilePath);

            DisplayMsg("Deleting file: " + strFilePath);
            if (!System.IO.File.Exists(strFilePath))
            {
                string strerr = "Error: path " + strFilePath + " did not exist for deletion";
                DisplayMsg(strerr);
                Globals.Logger.LogError(strerr);
            }
            else
            {
                try
                {
                    System.IO.File.Delete(strFilePath);
                }
                catch (Exception ex)
                {
                    string strErr = "Failed to delete file " + strFilePath + ":\r\n\r\n" + ex.ToString();
                    DisplayMsg(strErr);
                    Globals.Logger.LogError(strErr);
                    Exception ex2 = new Exception(strErr, ex);
                    throw ex2;
                }
            }
        }
        #endregion

        private void ExecuteUninstall()
        {
            InstallerFileTable objTable = null;
            InstallerBinary objBinary = null;
            InstallOptions objOptions = null;

            InstallState = InstallState.Uninstalling;

            //TODO: uninstaller has to store options AND uninstall table.
            try
            {
                // Read binary and stuff
                objBinary = new InstallerBinary();
                objBinary.OpenStream();
               // LoadOptions(); why can't we do this again?? idk
                objOptions = new InstallOptions();
                objOptions.Deserialize(objBinary, this);
                objTable = new InstallerFileTable(this);
                objTable.Deserialize(objBinary);
                objBinary.CloseStream();

                string strInstallRoot = objOptions.GetOptionValueOrDefault(InstallOption.UninstallFolderRoot_Uninstaller_Only);
                string strSavedDir = System.IO.Directory.GetCurrentDirectory();
                System.IO.Directory.SetCurrentDirectory(strInstallRoot);
                {
                    DisplayMsg("Executing Pre-Uninstall");
                    ExecutePreUninstall();

                    DisplayMsg("Executing Uninstall");
                    RollbackOrUninstall(objTable, strInstallRoot, 0.0);

                    DisplayMsg("Deleting Shortcuts");
                    DeleteShortcuts(objOptions.GetProgramInfo(), objOptions);//TODO: consolidate - 

                    DisplayMsg("Removing Registry Keys");
                    RemoveRegistryKeys(objOptions.GetProgramInfo());

                    DisplayMsg("Executing Post-Uninstall");
                    ExecutePostUninstall();

                    InstallState = InstallState.Successful;
                }
                System.IO.Directory.SetCurrentDirectory(strSavedDir);
            }
            catch (Exception ex)
            {
                InstallState = InstallState.Canceled;
                InstallErrors.Add("Error: " + ex.ToString());
                Globals.Logger.LogError(ex.ToString(), false, true);
            }
            finally
            {
                EndUninstall();
            }
        }
        private void DisplayMsg(string msg)
        {
            Globals.Logger.LogInfo(msg);
            _strCurrentFile = msg;
        }
        private void RollbackOrUninstall(InstallerFileTable objTable, string strInstallRoot, double initialProgress)
        {
            InstallState = InstallState.Rollingback;
            if (!System.IO.Directory.Exists(strInstallRoot))
            {
                Globals.Logger.LogError("Fatal Error: The original install directory '"
                    + strInstallRoot
                    + "' does not exist, is restricted, or was not packed in the uninstaller binary."
                    + " Installer will attempt to remove all remaining registry keys.", false, true);
            }
            else
            {
                if (objTable != null)
                    objTable.RemoveInstalledFiles(strInstallRoot, ref _dblProgress, ref _strCurrentFile, initialProgress);
            }
        }
        private string CreateInstallRoot()
        {
            string strInstallRoot = string.Empty;
            try
            {
                // - Checks
                Globals.Logger.LogInfo("Creating Root");
                InstallOption location = GetOption(InstallOption.InstallLocation);
                if (location == null)
                    Globals.Throw("Install location was not set.");

                strInstallRoot = ReplaceValueVariables(location.Value);

                Globals.Logger.LogInfo("Creating Root Dir..");
                if (!System.IO.Directory.Exists(strInstallRoot))
                    System.IO.Directory.CreateDirectory(strInstallRoot);
            }
            catch (Exception ex)
            {
                InstallErrors.Add("Error: " + ex.ToString());
                Globals.Logger.LogError(ex.ToString());
                return "";
            }
            return strInstallRoot;
        }
        private void EndInstallation()
        {
            if (InstallState == InstallState.Rollingback)
                InstallState = InstallState.Canceled;
            else if (InstallState == InstallState.Installing)
                InstallState = InstallState.Successful;
            else
                Globals.Throw("Install state was invalid for end installation.  The program may not have installed.");
        }
        private void EndUninstall()
        {
        }
        private void CreateUninstaller(ProgramInfo objProgramInfo, InstallerBinary objBinary, InstallerFileTable objTable, string strInstallRoot)
        {
            string strUninstallExePath;

            //Add the root to the options.
            _objOptions.AddOrReplaceOption(new InstallOption(InstallOption.UninstallFolderRoot_Uninstaller_Only, strInstallRoot, InstallerSectionType.UninstallSection));

            strUninstallExePath = objTable.WriteUninstaller(_objOptions, strInstallRoot);

            WriteUninstallerRegistryKeys(objProgramInfo, strUninstallExePath, strInstallRoot);
        }
        private void CreateAppRegistryKeys(ProgramInfo objProgramInfo, string strInstallRoot)
        {
            // Adds HKLM\Software\*Publisher*\*App*
            // Adds HKLM\Software\*Publisher*\*App*\InstallDir
            //**Note: registry keys in HKLM will fall under Wow6432Node in 32 bit apps. HKLM\Software\Wow6432Node\..
            RegistryKey softwareKey = Registry.LocalMachine.OpenSubKey(@"Software", true);

            if (!string.IsNullOrEmpty(objProgramInfo.Publisher))
            {
                RegistryKey publisherKey;
                publisherKey = softwareKey.OpenSubKey(objProgramInfo.Publisher, true);
                if (publisherKey == null)
                    publisherKey = softwareKey.CreateSubKey(objProgramInfo.Publisher);

                RegistryKey appKey;
                appKey = publisherKey.OpenSubKey(objProgramInfo.DisplayName, true);
                if (appKey == null)
                    appKey = publisherKey.CreateSubKey(objProgramInfo.DisplayName);

                appKey.SetValue("InstallDir", strInstallRoot);

                if (appKey != null)
                    appKey.Close();
                if (publisherKey != null)
                    publisherKey.Close();
            }
            else
                Globals.Logger.LogWarn("Program publisher was not set in the install config.  Cannot create /software registry key.");

            if (softwareKey != null)
                softwareKey.Close();
        }
        private RegistryKey GetUninstallerRegistryKey(ProgramInfo objProgramInfo, bool blnCreateIfNotFound = false)
        {
            RegistryKey appKey;
            RegistryKey softwareKey;

            softwareKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall", true);

            if (softwareKey == null)
                Globals.Logger.LogError("HKLM\\Software registry key not found.", true, true);

            string guidText = objProgramInfo.GetUninstallerGuidAsString();
            appKey = softwareKey.OpenSubKey(guidText, true) ?? softwareKey.CreateSubKey(guidText);

            if (appKey == null)
                Globals.Throw(String.Format("Unable to create uninstaller {0}", guidText));

            if (softwareKey != null)
                softwareKey.Close();

            return appKey;
        }
        private void WriteUninstallerRegistryKeys(ProgramInfo objProgramInfo, string strUninstallExePath, string installLocation)
        {
            //https://msdn.microsoft.com/en-us/library/aa372105(v=vs.85).aspx
            RegistryKey key = GetUninstallerRegistryKey(objProgramInfo, true);

            WriteUninstallRegistryKeyValues(objProgramInfo, key, strUninstallExePath, installLocation);
            if (key != null)
                key.Close();
        }
        private void WriteUninstallRegistryKeyValues(ProgramInfo objProgramInfo, RegistryKey key, string strUninstallExePath, string installLocation)
        {
            //**Note: if you modify this to add keys make sure to modify the RemoveInstallerRegistryKeys
            if (!String.IsNullOrEmpty(objProgramInfo.DisplayName))
                key.SetValue("DisplayName", objProgramInfo.DisplayName);
            else
            {
                Globals.Logger.LogWarn("DisplayName not set in installer configuration.");
                key.SetValue("DisplayName", "Spark Program");
            }

            key.SetValue("InstallLocation", installLocation);
            key.SetValue("ApplicationVersion", objProgramInfo.ApplicationVersion);
            key.SetValue("DisplayVersion", objProgramInfo.DisplayVersion);

            if (!String.IsNullOrEmpty(objProgramInfo.Publisher))
                key.SetValue("Publisher", objProgramInfo.Publisher);
            //  Icon.ExtractAssociatedIcon(Application.ExecutablePath);

            key.SetValue("DisplayIcon", strUninstallExePath);

            if (!string.IsNullOrEmpty(objProgramInfo.URLInfoAbout))
                key.SetValue("URLInfoAbout", objProgramInfo.URLInfoAbout);
            if (!string.IsNullOrEmpty(objProgramInfo.Contact))
                key.SetValue("Contact", objProgramInfo.Contact);
            if (!string.IsNullOrEmpty(objProgramInfo.Comments))
                key.SetValue("Comments", objProgramInfo.Comments);


            key.SetValue("InstallDate", DateTime.Now.ToString("yyyyMMdd"));
            key.SetValue("UninstallString", strUninstallExePath + " " + SparkFlags.Uninstall);

        }
        private void RemoveRegistryKeys(ProgramInfo objProgramInfo)
        {
            RemoveAppRegistryKeys(objProgramInfo);
            RemoveInstallerRegistryKeys(objProgramInfo);
        }
        private void RemoveAppRegistryKeys(ProgramInfo objProgramInfo)
        {
            // ** This keeps the publisher key there. I tis possible to remove publisher key
            // if there are no further keys.. but we don't do that.
            RegistryKey softwareKey = Registry.LocalMachine.OpenSubKey(@"Software", true);
            if (softwareKey != null)
            {
                RegistryKey publisherKey;
                publisherKey = softwareKey.OpenSubKey(objProgramInfo.Publisher, true);
                if (publisherKey != null)
                {
                    publisherKey.DeleteSubKeyTree(objProgramInfo.DisplayName, false);
                    publisherKey.Close();
                }
                else
                    Globals.Logger.LogError("Could not find application subkey " + objProgramInfo.DisplayName + ".  The uninstall may not have succeeded.");
                softwareKey.Close();
            }
            else
                Globals.Logger.LogError("Could not find HKLM\\Software.  The uninstall may not have succeeded.");
        }
        private void RemoveInstallerRegistryKeys(ProgramInfo objProgramInfo)
        {
            RegistryKey softwareKey;

            string guidText = objProgramInfo.GetUninstallerGuidAsString();

            softwareKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall", true);

            if (softwareKey == null)
            {
                Globals.Logger.LogError("HKLM\\Software registry key not found.");
                return;
            }

            softwareKey.DeleteSubKeyTree(guidText, false);

            if (softwareKey != null)
                softwareKey.Close();
        }

        private void CreateShortcuts(ProgramInfo objProgramInfo, InstallOptions objOptions)
        {
            //Creates all shortcuts (start menu / desktop / etc)
            foreach(ShortcutOption opt in _objOptions.GetShortcutOptions())
            {
                if (opt.Enabled)
                    InstallShortcutOption(opt);
                else
                    Globals.Logger.LogInfo("User deselected option " + opt.ShortcutName);
            }
        }
        private void InstallShortcutOption(ShortcutOption sc)
        {
            //Warning:Display Name may have spaces.

            DisplayMsg("Creating Shortcut at " + sc.ShortcutPath);
            DisplayMsg("  Name " + sc.ShortcutName);
            DisplayMsg("  Icon Path " + sc.IconPath);
            DisplayMsg("  Target Path " + sc.TargetPath);
            DisplayMsg("  Args " + sc.Arguments);
            DisplayMsg("  Working Dir " + sc.WorkingDirectory);
            // Save the shortcut
            WshShell shell = new WshShell();
            IWshShortcut shortcut = (IWshShortcut)shell.CreateShortcut(sc.ShortcutPath);
            shortcut.Description = sc.ShortcutName;
            //shortcut.IconLocation = @"C:\Program Files (x86)\TestApp\TestApp.ico"; //uncomment to set the icon of the shortcut
            if(!string.IsNullOrEmpty(sc.IconPath))
                shortcut.IconLocation = sc.IconPath;
            shortcut.TargetPath = sc.TargetPath;
            if(!string.IsNullOrEmpty(sc.Arguments))
                shortcut.Arguments = sc.Arguments;
            shortcut.WorkingDirectory = sc.WorkingDirectory;
            shortcut.Save();
            DisplayMsg("Done.");

        }
        private void DeleteShortcuts(ProgramInfo objProgramInfo, InstallOptions objOptions)
        {

            // * delete all shortcut objects first
            foreach (ShortcutOption opt in _objOptions.GetShortcutOptions())
            {
                DisplayMsg("  Deleting Shortcut:" + opt.ShortcutPath);
                if (System.IO.File.Exists(opt.ShortcutPath))
                    System.IO.File.Delete(opt.ShortcutPath);
                else
                    Globals.Logger.LogWarn("Could not find or delete shortcut File " + opt.ShortcutPath);
            }
            // * next delete all folders.
            //**UNSAFE - this will try to delete the desktop, etc.
            foreach (ShortcutOption opt in _objOptions.GetShortcutOptions())
            {
                string dir = System.IO.Path.GetDirectoryName(opt.ShortcutPath);

                DisplayMsg("  Deleting Shortcut Dir:" + dir);
                if (System.IO.Directory.Exists(dir) 
                    && System.IO.Directory.GetFiles(dir).Length==0 
                    && System.IO.Directory.GetDirectories(dir).Length==0)
                    System.IO.Directory.Delete(dir);
                else
                    Globals.Logger.LogWarn("  Could not find or delete shortcut Path " + dir);
            }
        }

        #endregion

    }
}
